/*
 * 
 * Date:        2008/10/24
 * Info:        mod_slotlimit Apache 2 module
 * Contact:     mailto: <luca.e@seeweb.it>
 * Version:     1.1 
 * Author: 	Luca Ercoli
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *
 */


#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_protocol.h"
#include "http_core.h"
#include "http_main.h"
#include "http_log.h"
#include "ap_mpm.h"
#include "apr_strings.h"
#include "scoreboard.h"

#define VHOST_LEN     32 

#define MODULE_NAME "mod_slotlimit"
#define MODULE_VERSION "1.1"

module AP_MODULE_DECLARE_DATA slotlimit_module;

static void *slotlimit_create_config(apr_pool_t *p, server_rec *s);
static void *slotlimit_create_dir_config(apr_pool_t *p, char *path);
static int slotlimit_handler(request_rec *r);
static const char *limit_config_cmd(cmd_parms *parms, void *dummy,int arg);
static const char *limit_config2_cmd(cmd_parms *parms, void *dummy,int arg);
static const char *limit_config3_cmd(cmd_parms *parms, void *dummy,int arg);
static const char *limit_config4_cmd(cmd_parms *parms, void *dummy,char *arg);
static const char *limit_config5_cmd(cmd_parms *parms, void *dummy,int arg);
static const char *force_vhost_name(cmd_parms *cmd, void *dummy, int arg);
static int slotlimit_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s);
static void register_hooks(apr_pool_t *p);

char temp[VHOST_LEN];

static int server_limit, thread_limit;


typedef struct {
	int conn_num;
	int site_conn_num;
	int perc_slots_available;
	char *sitename_to_limit;
	char *errore;
	int client_ip;
	int conn_limitsite;
	int is_set_vhost;
} mod_config;


static mod_config *create_config(apr_pool_t *p)
{
	mod_config *cfg = (mod_config *)
		apr_pcalloc(p, sizeof (*cfg));

	cfg->conn_num = 65432;
	cfg->site_conn_num = 65432;
	cfg->perc_slots_available = 13;
	cfg->sitename_to_limit = "LucaErcoli";
	cfg->client_ip=65432;
	cfg->conn_limitsite=20;
	cfg->errore="Blocked by mod_slotlimit. More information about this error may be available in the server error log.";
	return cfg;
}




/* per-server configuration structure */
static void *slotlimit_create_config(apr_pool_t *p, server_rec *s)
{
	return create_config(p);
}



/* per-directory configuration structure */
static void *slotlimit_create_dir_config(apr_pool_t *p, char *path)
{
	return create_config(p);
}



static int core(request_rec *r, mod_config *cfg)
{

	if (!ap_is_initial_req(r))
	{
		return DECLINED;
	}


	int i;
	int j;


	int clientip_count = 0;
	int vhost_count = 0;
	int single_count = 0;
	int nAllSlots = 0;        

	char siti_visitati[2048][VHOST_LEN];


	worker_score *ws_record;
	

	if (cfg->is_set_vhost)
	{
		ws_record = ap_get_scoreboard_worker(r->connection->id, 0);

		/*
		 * function ap_update_child_status (defined in server/scoreboard.c) overwrite ws->vhost
		 *
		 */
		/* According to RFC 2616, Host header field CAN be blank. */	


	if ( apr_table_get(r->headers_in,"Host"))
	{
	memset(temp,'\0',VHOST_LEN);
	strncpy(temp,apr_table_get(r->headers_in,"Host"),VHOST_LEN);

	r->server->server_hostname = &temp;
	
	conn_rec *c = r->connection;
	(void)ap_update_child_status(c->sbh, SERVER_BUSY_WRITE, r);

	}


	}




	for (i = 0; i < server_limit; ++i) {
		for (j = 0; j < thread_limit; ++j) {
			ws_record = ap_get_scoreboard_worker(i, j);
			switch (ws_record->status) {
				case SERVER_BUSY_READ:
				case SERVER_BUSY_WRITE:
				case SERVER_BUSY_KEEPALIVE:
				//case SERVER_BUSY_LOG:
				//case SERVER_BUSY_DNS:


					strncpy(siti_visitati[nAllSlots],ws_record->vhost,VHOST_LEN);

					nAllSlots++;


					if ( (strcmp(r->server->server_hostname, ws_record->vhost) == 0)  ) vhost_count++;

					if ( (strcmp(cfg->sitename_to_limit, ws_record->vhost) == 0) ) single_count++;

					if (strcmp(r->connection->remote_ip,ws_record->client) == 0) clientip_count++;


					break;

				default:
					break;
			}
		}
	}



	if ( (vhost_count > cfg->site_conn_num) ){
		ap_log_rerror(APLOG_MARK, APLOG_CRIT, 0, r,
				"mod_slotlimit: connection denied! Host %s reached MaxConnectionsPerSite (%d)",r->hostname,cfg->site_conn_num);
		ap_custom_response(r, HTTP_NOT_ACCEPTABLE, cfg->errore);
		return HTTP_NOT_ACCEPTABLE;
	}


	if ((single_count > cfg->conn_limitsite) && (strcmp (cfg->sitename_to_limit,r->hostname) == 0) ){
		ap_log_rerror(APLOG_MARK, APLOG_CRIT, 0, r,
				"mod_slotlimit: connection denied! Host %s reached LimitSiteConnections (%d)",r->hostname,cfg->conn_limitsite);
		ap_custom_response(r, HTTP_NOT_ACCEPTABLE, cfg->errore);
		return HTTP_NOT_ACCEPTABLE;
	}


	if (clientip_count > cfg->client_ip){

		ap_log_rerror(APLOG_MARK, APLOG_CRIT, 0, r,
				"mod_slotlimit: connection denied! Client %s reached ClientIpLimit (%d) Site: %s",r->connection->remote_ip,cfg->client_ip,r->hostname);
		ap_custom_response(r, HTTP_NOT_ACCEPTABLE, cfg->errore);
		return HTTP_NOT_ACCEPTABLE;

	}



	if ( nAllSlots >= (cfg->conn_num - ( (cfg->conn_num * cfg->perc_slots_available) / 100)) ){	


		qsort ( (char *)siti_visitati, nAllSlots, sizeof (*siti_visitati), strcmp);

		j = 0;
		i = 0;

		for (i; i < nAllSlots ; i++)  if (strcmp(siti_visitati[i+1],siti_visitati[i])) j++;


		if ( vhost_count >= ( (cfg->conn_num / j) + (cfg->conn_num - nAllSlots)  ) ){

			ap_log_rerror(APLOG_MARK, APLOG_CRIT, 0, r, "mod_slotlimit: limit incoming connections! SiteName: %s  SlotsUsed: %d  TotalSlotsBusy: %d  MaxClients: %d LimitTo %d DifferentSitesVisited %d",r->hostname,vhost_count,nAllSlots,cfg->conn_num,((cfg->conn_num / j) + (cfg->conn_num - nAllSlots)),j );

			ap_custom_response(r, HTTP_NOT_ACCEPTABLE, cfg->errore);
			return HTTP_NOT_ACCEPTABLE;
		} 

	}

	return DECLINED;
}


static int slotlimit_handler(request_rec *r)
{
	mod_config *cfg = (mod_config *)
		ap_get_module_config(r->per_dir_config, &slotlimit_module);

	int result;

	result = core(r, cfg);

	return result;
}


static const char *limit_config_cmd(cmd_parms *parms, void *dummy,
		int arg)
{

	mod_config *cfg = (mod_config *)dummy; 
	ap_get_module_config(parms->server->module_config, &slotlimit_module);

	int num_req = atoi (arg);
	cfg->conn_num = num_req; 
	return NULL;
}

static const char *limit_config2_cmd(cmd_parms *parms, void *dummy,
		int arg)
{

	mod_config *cfg = (mod_config *)dummy;
	ap_get_module_config(parms->server->module_config, &slotlimit_module);

	int num_req = atoi (arg);
	cfg->site_conn_num = num_req;
	return NULL;
}



static const char *limit_config3_cmd(cmd_parms *parms, void *dummy,
		int arg)
{

	mod_config *cfg = (mod_config *)dummy;
	ap_get_module_config(parms->server->module_config, &slotlimit_module);

	int num_req = atoi (arg);
	cfg->perc_slots_available = num_req;
	return NULL;
}

static const char *limit_config4_cmd(cmd_parms *parms, void *dummy,
		char *arg)
{

	mod_config *cfg = (mod_config *)dummy;
	ap_get_module_config(parms->server->module_config, &slotlimit_module);

	char *qualesito = arg;
	cfg->sitename_to_limit = qualesito;
	return NULL;
}


static const char *limit_config7_cmd(cmd_parms *parms, void *dummy,
                char *arg)
{

        mod_config *cfg = (mod_config *)dummy;
        ap_get_module_config(parms->server->module_config, &slotlimit_module);

        char *ops = arg;
        cfg->errore = ops;
        return NULL;
}


static const char *limit_config5_cmd(cmd_parms *parms, void *dummy,
		int arg)
{

	mod_config *cfg = (mod_config *)dummy;
	ap_get_module_config(parms->server->module_config, &slotlimit_module);

	int num_req = atoi (arg);
	cfg->client_ip = num_req;
	return NULL;
}

static const char *limit_config6_cmd(cmd_parms *parms, void *dummy,
		int arg)
{

	mod_config *cfg = (mod_config *)dummy;
	ap_get_module_config(parms->server->module_config, &slotlimit_module);

	int num_req = atoi (arg);
	cfg->conn_limitsite = num_req;
	return NULL;
}



static const char *force_vhost_name(cmd_parms *cmd, void *dummy, int arg)
{
	const char *err = ap_check_cmd_context(cmd, GLOBAL_ONLY);
	if (err != NULL) {
		return err;
	}

	mod_config *cfg = (mod_config *)dummy;
	ap_get_module_config(cmd->server->module_config, &slotlimit_module);


	cfg->is_set_vhost = arg;
	return NULL;
}



static command_rec slotlimit_cmds[] = {

	AP_INIT_TAKE1("MaxClients", limit_config_cmd, NULL, RSRC_CONF,"Apache maxclient setting"),
	AP_INIT_TAKE1("AvailableSlotsPercent", limit_config3_cmd, NULL, RSRC_CONF,"Percentage of slots available in order to set any restrictions"),
	AP_INIT_TAKE1("MaxConnectionsPerSite", limit_config2_cmd, NULL, RSRC_CONF,"Max connections for each running site"),
	AP_INIT_TAKE1("LimitSite", limit_config4_cmd, NULL, RSRC_CONF,"Single site to limit"),
	AP_INIT_TAKE1("LimitSiteConnections", limit_config6_cmd, NULL, RSRC_CONF,"Max connections for LimitSite"),
	AP_INIT_TAKE1("ClientIpLimit", limit_config5_cmd, NULL, RSRC_CONF,"Limit of simultaneous connection per IP"),
	AP_INIT_TAKE1("CustomErrMsg", limit_config7_cmd, NULL, RSRC_CONF,"Custom error message"),
	AP_INIT_FLAG("ForceVhostName", force_vhost_name, NULL, RSRC_CONF,"\"On\" to force vhost hostname in scoreboard, \"Off\" to disable"),
	{NULL}
};

static int slotlimit_init(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
			MODULE_NAME " " MODULE_VERSION " started.");
	ap_mpm_query(AP_MPMQ_HARD_LIMIT_THREADS, &thread_limit);
	ap_mpm_query(AP_MPMQ_HARD_LIMIT_DAEMONS, &server_limit);
	return OK;
}

static void register_hooks(apr_pool_t *p)
{

	ap_hook_post_config(slotlimit_init, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_access_checker(slotlimit_handler,NULL,NULL,APR_HOOK_MIDDLE);

}


module AP_MODULE_DECLARE_DATA slotlimit_module = {
	STANDARD20_MODULE_STUFF,
	slotlimit_create_dir_config, /* create per-dir config structures */
	NULL,                       /* merge  per-dir    config structures */
	slotlimit_create_config,  /* create per-server config structures */
	NULL,                       /* merge  per-server config structures */
	slotlimit_cmds,           /* table of config file commands       */
	register_hooks
};
